"""
PieTunes is an abstraction of Apple's Scripting Bridge API for iTunes.
Copyright (C) 2018  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""


from enum import IntEnum


class iTunesESrA(IntEnum):
    ALBUMS = 0x6b53724c
    ALL = 0x6b416c6c
    ARTISTS = 0x6b537252
    COMPOSERS = 0x6b537243
    DISPLAYED = 0x6b537256
    SONGS = 0x6b537253



"""
PieTunes is an abstraction of Apple's Scripting Bridge API for iTunes.
Copyright (C) 2018  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

from collections import namedtuple

Version = namedtuple(
    '_version_tuple',
    ['major', 'minor', 'micro', 'pre_tags', 'post_tag']
)

version_pre_tags = []
version_post_tag = ''
version_git_tags = []
version_git_rev = ''


def _basic_version(version):
    """Creates a basic, semantic version string

       For additional information on Semantic Versioning,
       see https://semver.org

        Args:
            version: An instance of a Version namedtuple

        Returns:
            A string created from the Version tuple, formatted using
            basic Smeantic Versioning.
            example:

            '1.2.3'
    """
    v = '.'.join([str(getattr(version, f'{field}'))
                  for field in version._fields
                  if field in
                  ('major', 'minor', 'micro')])

    return v


def _version_to_git_string(version):
    pass


def _version_to_string(version):
    """Creates a version string according to PEP-440

       For additional information, please
       see https://www.python.org/dev/peps/pep-0440/#version-scheme

        Args:
            version: An instance of a Version namedtuple

        Returns:
            A string created from the Version tuple, formatted according
            to PEP-440
            example:

            '0.0.1.dev.b1'
    """
    s1 = '.' if version.pre_tags else ''
    s2 = '.' if version.pre_tags else ''
    v = f'.'.join([str(getattr(version, f'{field}'))
                   for field in version._fields
                   if field in ('major', 'minor', 'micro')]) + \
        f'{s1}' + f'.'.join([tag for tag in version.pre_tags]) + \
        f'{s2}{version.post_tag}'

    return v


# To increment the version, increment it in the Version tuple _version_t_
_version_t_ = Version(0, 0, 3, version_pre_tags, version_post_tag)
_short_version_ = _basic_version(_version_t_)
__version__ = _version_to_string(_version_t_)

_version_min_python_t_ = Version(3, 6, 0, [], '')
_version_min_python_ = _version_to_string(_version_min_python_t_)



"""
PieTunes is an abstraction of Apple's Scripting Bridge API for iTunes.
Copyright (C) 2018  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""


from types import SimpleNamespace

from ScriptingBridge import SBApplication

from pietunes import core


DEFAULT_LIBRARY_LOCATION = "com.apple.iTunes 'book:1:iTunes Library Location'"


class App(object):
    """The App class provides a Python API for scripting iTunes

    The client application built on this pietunes package should
    instantiate an instance of this class in order to acces the API.
    All messages sent to iTunes should use only the methods provided here.

    Attributes:
        itunes: An SBApplication instance for iTunes.app
        playlists: A dynamically updated list of all playlists known by iTunes
    """

    def __init__(self):
        """Inits App with SBAppication and a list of available playlists."""
        self.itunes = SBApplication.applicationWithBundleIdentifier_(
            "com.apple.iTunes")
        self.playlists = self._get_playlists()

    def _get_playlists(self):
        return list(self.itunes.playlists())

    def get_playlist(self, name):
        """Return a reference to an iTunes Playlist object.

        Args:
            playlist (str): the name of the iTunes playlist

        Returns:
            an iTunes Playlist object
        """
        playlist = None
        candidate = iter(self.playlists)
        while not playlist:
            try:
                p = next(candidate)
                if p.name() == name:
                    playlist = p
                    return playlist
            except StopIteration:
                raise NotFoundError(f'\nNo playlist with the name {name}'
                                    ' could be found.\n')

    def get_sources(self):
        """Returns a list of iTunes data sources.
        """
        return self.itunes.sources()

    def get_tracks(self, playlist):
        """Retrieve the iTunes track objects from the playlist refernce

        Args:
            playlist (str): the name of the playlist that collects the tracks

        Returns:
            an iterator object that yields iTunes track objects
        """
        tracks = (track for track in playlist.tracks())

        return tracks

    def get_track(self, collection, title):
        """Returns an a reference to an iTunes Track object.

        Args:
            collection (generator): a generator object
                                    that yields track objects
            title (str): the title of the track to retrieve

        Returns:
            an iTunes track object
        """
        track = None
        candidate = iter(collection)
        while not track:
            try:
                t = next(candidate)
                if t.name() == title:
                    track = t
                    return track
            except StopIteration:
                raise NotFoundError(f'\nNo movie with the title {title}'
                                    ' could be found.\n')

    def download_track(self, playlist, track):
        """Downloads an iTunes track.

        Args:
            playlist (str): the name of the playlist that contains the track
            track (str): the name of the track to download

        """
        nsargs = SimpleNamespace(
            script='download_track',
            playlist=playlist,
            track=track
        )
        core._call_script(nsargs)

    def open_library(self, location):
        """Open an arbitrary iTunes Library.

        Args:
            location (str): the iTunes Bookmark of the iTunes Library location

        """
        nsargs = SimpleNamespace(script='open_library', location=location)
        core._call_script(nsargs)

    def search(self, playlist, search_text, constraint):
        """Search a playlist for a string, subject to a constraint on the type
        of object to search for.

        Args:
            playlist (str): the name of the iTunes playlist to search
            search_text (str): The text string to search for
            constraint (enum): The subset of objects in the playlist to search
                May be one of:
                    ALBUMS

                    ALL

                    ARTISTS

                    COMPOSERS

                    DISPLAYED

                    SONGS

        Returns:
            a list of objects that at least partially match the search string
        """
        target = self.get_playlist(playlist)
        matches = target.searchFor_only_(search_text, constraint)

        return matches


class NotFoundError(Exception):
    """Error thrown when the item being searched for is not found.
    """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value



"""
PieTunes is an abstraction of Apple's Scripting Bridge API for iTunes.
Copyright (C) 2018  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""


import subprocess

from pietunes.script_support import _download, _open_library


def _call_script(nsargs):
    if nsargs.script == 'download':
        script = _download(nsargs.playlist, nsargs.track)
    elif nsargs.script == 'open_library':
        script = _open_library(nsargs.location)

    osa = subprocess.Popen(
        ['osascript', '-'],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    osa.communicate(script)



"""
PieTunes is an abstraction of Apple's Scripting Bridge API for iTunes.
Copyright (C) 2018  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

from setuptools import setup, find_packages

from pietunes.__version__ import __version__, _version_min_python_

with open("README.rst", "r") as fh:
    long_description = fh.read()

required = [
    "pyobjc>=5.1.1",
]

setup(
    name='pietunes',
    version=__version__,
    python_requires=f">={_version_min_python_}",
    install_requires=required,

    packages=find_packages(),

    license='AGPLv3',
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "License :: OSI Approved :: "
        "GNU Affero General Public License v3 or later (AGPLv3+)",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Multimedia :: Video",
        "Topic :: Software Development :: Libraries",
        "Environment :: Console",
        "Operating System :: MacOS",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
    ],

    url='https://pypi.org/project/pietunes/',
    author="Brian Farrell",
    author_email="brian.farrell@me.com",
    description="A library to aid in automating iTunes via ScriptingBridge.",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    keywords="modern development iTunes ScriptingBridge osx darwin",
    zip_safe=False,
)



"""
PieTunes is an abstraction of Apple's Scripting Bridge API for iTunes.
Copyright (C) 2018  Brian Farrell

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact: brian.farrell@me.com
"""

from textwrap import dedent


def _download(requested_playlist, requested_track):
    script = f"""
    tell application "iTunes"
        set playlist_target to some playlist ¬
            whose name is "{requested_playlist}"
        set len to count of tracks of playlist_target
        set target_collection to (tracks of playlist_target)
        repeat with i from 1 to number of items in target_collection
            set candidate to item i of target_collection
            if (name of candidate) = "{requested_track}" then
                download candidate
                exit repeat
            end if
        end repeat
    end tell
    """
    return dedent(script).encode('utf-8')


def _open_library(library_location):
    """launch iTunes with a given Library
    This script came from https://stackoverflow.com/a/1693973
    Courtesy of Nicholas Riley:
        https://stackoverflow.com/users/6372/nicholas-riley
    """

    script = f"""
    property altLibraryLocation : "{library_location}"
    property libraryLocationPref : ¬
        "com.apple.iTunes 'book:1:iTunes Library Location'"



    """
    return dedent(script).encode('utf-8')


def _quit_itunes():
    script = f"""
    tell application "System Events"
        if exists (application process "iTunes") then
            tell application "iTunes" to quit
        end if
    end tell
    """

    return dedent(script).encode('utf-8')
